# test_llama_index.py
import pytest
from tokencost.callbacks import llama_index
from llama_index.core.callbacks.schema import CBEventType, EventPayload
from unittest.mock import MagicMock

# Mock the calculate_prompt_cost and calculate_completion_cost functions

# 4 tokens
STRING = "Hello, world!"


# Mock the ChatMessage class in LlamaIndex
@pytest.fixture
def mock_chat_message(monkeypatch):
    class MockChatMessage:
        def __init__(self, text):
            self.text = text

        def __str__(self):
            return self.text

    monkeypatch.setattr("llama_index.core.llms.ChatMessage", MockChatMessage)
    return MockChatMessage


# Test the _calc_llm_event_cost method for prompt and completion


def test_calc_llm_event_cost_prompt_completion(capsys):
    handler = llama_index.TokenCostHandler(model="gpt-3.5-turbo")
    payload = {EventPayload.PROMPT: STRING, EventPayload.COMPLETION: STRING}
    handler._calc_llm_event_cost(payload)
    captured = capsys.readouterr()
    assert "# Prompt cost: 0.0000060" in captured.out
    assert "# Completion: 0.000008" in captured.out


# Test the _calc_llm_event_cost method for messages and response


def test_calc_llm_event_cost_messages_response(mock_chat_message, capsys):
    handler = llama_index.TokenCostHandler(model="gpt-3.5-turbo")
    payload = {
        EventPayload.MESSAGES: [
            mock_chat_message("message 1"),
            mock_chat_message("message 2"),
        ],
        EventPayload.RESPONSE: "test response",
    }
    handler._calc_llm_event_cost(payload)
    captured = capsys.readouterr()
    assert "# Prompt cost: 0.0000105" in captured.out
    assert "# Completion: 0.000004" in captured.out


# Additional tests can be written for start_trace, end_trace, on_event_start, and on_event_end
# depending on the specific logic and requirements of those methods.



import os
import json
import aiohttp
import asyncio
import logging

"""
Prompt (aka context) tokens are based on number of words + other chars (eg spaces and punctuation) in input.
Completion tokens are similarly based on how long chatGPT's response is.
Prompt tokens + completion tokens = total tokens.
The max total limit is typically 1 more than the prompt token limit, so there's space for at least one completion token.

You can use ChatGPT's webapp (which uses their tiktoken repo) to see how many tokens your phrase is:
https://platform.openai.com/tokenizer

Note: When asking follow-up questions, everything above and including your follow-up question
is considered a prompt (for the purpose of context) and will thus cost prompt tokens.
"""

# How to read TOKEN_COSTS:
# Each prompt token costs __ USD per token.
# Each completion token costs __ USD per token.
# Max prompt limit of each model is __ tokens.

PRICES_URL = "https://raw.githubusercontent.com/BerriAI/litellm/main/model_prices_and_context_window.json"


async def fetch_costs():
    """Fetch the latest token costs from the LiteLLM cost tracker asynchronously.
    Returns:
        dict: The token costs for each model.
    Raises:
        Exception: If the request fails.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(PRICES_URL) as response:
            if response.status == 200:
                return await response.json(content_type=None)
            else:
                raise Exception(f"Failed to fetch token costs, status code: {response.status}")


async def update_token_costs():
    """Update the TOKEN_COSTS dictionary with the latest costs from the LiteLLM cost tracker asynchronously."""
    global TOKEN_COSTS
    try:
        TOKEN_COSTS = await fetch_costs()
    except Exception as e:
        logging.error(f"Failed to update TOKEN_COSTS: {e}")

with open(os.path.join(os.path.dirname(__file__), "model_prices.json"), "r") as f:
    TOKEN_COSTS_STATIC = json.load(f)


# Ensure TOKEN_COSTS is up to date when the module is loaded
try:
    asyncio.run(update_token_costs())
except Exception:
    logging.error('Failed to update token costs. Using static costs.')
    TOKEN_COSTS = TOKEN_COSTS_STATIC



from typing import Any, Dict, List, Optional, cast
from llama_index.core.callbacks.base_handler import BaseCallbackHandler
from llama_index.core.callbacks.schema import CBEventType, EventPayload
from llama_index.core.llms import ChatMessage
from tokencost import calculate_all_costs_and_tokens


class TokenCostHandler(BaseCallbackHandler):
    """Callback handler for printing llms inputs/outputs."""

    def __init__(self, model) -> None:
        super().__init__(event_starts_to_ignore=[], event_ends_to_ignore=[])
        self.model = model
        self.prompt_cost = 0
        self.completion_cost = 0
        self.prompt_tokens = 0
        self.completion_tokens = 0

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return

    def _calc_llm_event_cost(self, payload: dict) -> None:
        if EventPayload.PROMPT in payload:
            prompt = str(payload.get(EventPayload.PROMPT))
            completion = str(payload.get(EventPayload.COMPLETION))
            estimates = calculate_all_costs_and_tokens(prompt, completion, self.model)

        elif EventPayload.MESSAGES in payload:
            messages = cast(List[ChatMessage], payload.get(EventPayload.MESSAGES, []))
            messages_str = "\n".join([str(x) for x in messages])
            response = str(payload.get(EventPayload.RESPONSE))
            estimates = calculate_all_costs_and_tokens(
                messages_str, response, self.model
            )

        self.prompt_cost += estimates["prompt_cost"]
        self.completion_cost += estimates["completion_cost"]
        self.prompt_tokens += estimates["prompt_tokens"]
        self.completion_tokens += estimates["completion_tokens"]

        print(f"# Prompt cost: {estimates['prompt_cost']}")
        print(f"# Completion: {estimates['completion_cost']}")
        print("\n")

    def reset_counts(self) -> None:
        self.prompt_cost = 0
        self.completion_cost = 0
        self.prompt_tokens = 0
        self.completion_tokens = 0

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> str:
        return event_id

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """Count the LLM or Embedding tokens as needed."""
        if event_type == CBEventType.LLM and payload is not None:
            self._calc_llm_event_cost(payload)
